# vi:fdm=marker fdl=0:
# $Id: DBF.pm,v 1.45 2003/02/08 17:16:58 jettero Exp $ 

package FoxPro::DBF;

use strict;
use warnings;
use Carp;
use Fcntl qw(:seek);
use Date::Manip;

our $VERSION = 0.16;

# new and DESTROY {{{
sub new {
    my $this = shift;

    $this = bless {
        file    => shift,
        cur_row => 0,
    }, $this;

    croak "that's not a dbf file"        unless $this->{file} =~ m/dbf$/i;
    croak "no such file '$this->{file}'" unless -f $this->{file};

    $this->{table_name} = lc($1) if $this->{file} =~ m/([^\/]+)\.dbf$/i;

    croak "couldn't determin table name from \"$this->{file}\"" 
        unless $this->{table_name} and $this->{file} =~ m/\Q$this->{table_name}\E/;

    open $this->{IN}, "$this->{file}" or croak "couldn't open $this->{file}: $!";

    $this->read_header;

    return $this;
}

sub DESTROY { 
    my $this = shift;

    close $this->{IN};
}
# }}}
# read_header {{{
sub read_header {
    my $this = shift;

    my $header;
    seek($this->{IN}, 0, SEEK_SET) or croak "bad seek: $!";
    read($this->{IN}, $header, 32) == 32 
        and $header =~ m/^[\x30\xf5](.)(.)(.)(.{4})(..)(..)/ms 
         or die "doesn't feel like a foxpro 5 table";

    $this->{header}->{records}      = unpack("V",$4);
    $this->{header}->{header_bytes} = unpack("v",$5);
    $this->{header}->{record_bytes} = unpack("v",$6);
    $this->{header}->{last_update}  = 
    &UnixDate( 
        &ParseDate( 
            sprintf('%02d-%02d-%02d', unpack("C", $1), unpack("C", $2), unpack("C", $3))), '%Y-%m-%d');

    my $buf; $this->{field_order} = [];
    my $xxx; 
    my $so_far = 32 + 1; # the stuff from above -- plus an extra one, cuz we start counting from 0.
    while( 32 == read($this->{IN}, $buf, 32) and not $buf =~ m/^\x0d\x00{31}$/ms and $so_far < $this->{header}->{header_bytes} ) {
        my ($field, $typec) = (lc($1), $2) if $buf =~ m/^(\w{1,10})\x00+(.)/ms;

        my $type = {
            "C" => "char",
            "D" => "date",
            "T" => "datetime",
            "N" => "double",
            "F" => "float",
            "I" => "int",
            "B" => "bigint",
            
            "Y" => "double",         # currency
            "L" => "enum('t', 'f')", # logical
            "M" => "text",           # memo
            "G" => "blob",           # general

            "0" => undef, # the nullfields
        }->{$typec};

        my $disp  = unpack("V", $1) if $buf =~ m/^.{12}(....)/ms;
        my $width = unpack("C", $1) if $buf =~ m/^.{16}(.)/ms;
        my $decpl = unpack("C", $1) if $buf =~ m/^.{17}(.)/ms;
        my $null  = 0; $null = int(++$xxx) if $buf =~ m/^.{18}(\x02)/ms;

            $this->{fields}->{$field}->{type_c}         = $typec;
            $this->{fields}->{$field}->{width}          = $width;
            $this->{fields}->{$field}->{decimal_places} = $decpl if $decpl;
            $this->{fields}->{$field}->{displacement}   = $disp;
            $this->{fields}->{$field}->{null}           = $null;

            if( $type ) {
                my $mysql = lc("\`$field\`") . " $type";
                   $mysql .= "($width)"         if $typec =~ m/[cib]/ims;
                   $mysql .= "($width, $decpl)" if $typec =~ m/[nfy]/ims;
                   $mysql .= " ";
                   $mysql .= ($null ? "" : "not null");

                   $this->{fields}->{$field}->{mysql} = $mysql;
                   $this->{fields}->{$field}->{type}  = $type;
            }

            push @{ $this->{field_order} }, $field;

        if( $typec =~ m/[gm]/ims and not $this->{f_type_warn}->{$typec} = 1) {
            $this->{f_type_warn}->{$typec} = 1;
            carp "field type $typec (aka $type) is unsupported at this time";
        }

        $so_far += 32;
    }
}
# }}}
# mysql table create {{{
sub mysql_table_create {
    my $this = shift;
    my @rows = map( $this->{fields}->{$_}->{mysql}, grep(defined($this->{fields}->{$_}->{mysql}), @{ $this->{field_order} }) );
    my $rows = join( ",\n\t", @rows );

    my $temp = (shift() ? " temporary" : "");

    return qq/create$temp table `$this->{table_name}` (\n\t$rows\n)/;
}
# }}}
# mysql table drop {{{
sub mysql_table_drop { 
    my $this = shift;

    return "drop table if exists `$this->{table_name}`";
}
# }}}
# hash dump {{{
sub hash_dump {
    my $this = shift;
    my $file = shift;

    my %hash = %$this;

    delete $hash{IN};
    delete $hash{cur_row};

    use Data::Dumper;

    croak "You cannot mean what you're saying, you'll stomp on the table ... " if $file eq $this->{file};

     open OUT_DUMP, ">$file" or croak "can't open $file: $!";
    print OUT_DUMP  "# ", "v", "i:", "syntax=perl:\n\n";  # fools vi into ignoreing the commands...
    print OUT_DUMP  Dumper( \%hash );
    close OUT_DUMP;
}
# }}}
# fetchrow_hashref_by_n {{{
sub fetchrow_hashref_by_n {
    my $this = shift;
    my $row  = shift || 0;
    my $pos  = $this->{header}->{header_bytes} + $this->{header}->{record_bytes} * $row;

    my $data;
    seek($this->{IN}, $pos, SEEK_SET) or croak "bad seek($row, $pos): $!";
    read($this->{IN}, $data, $this->{header}->{record_bytes}) == $this->{header}->{record_bytes}
        or return undef;

    my $hasnul = 0;
    my @fields = @{ $this->{field_order} };

    if( $fields[$#fields] =~ m/_nullflags/) {
        my $field = pop @fields;

        $hasnul = unpack("C", substr($data, 
            $this->{fields}->{$field}->{displacement},
            $this->{fields}->{$field}->{width},
        ));
    }

    my $h = {};
    foreach my $field (@fields) {
        my $typec = $this->{fields}->{$field}->{type_c};

        $h->{__deleted_row} = substr($data, 0, 1);
        $h->{__deleted_row} = 1 if $h->{__deleted_row} eq "*";
        $h->{__deleted_row} = 0 if $h->{__deleted_row} eq " ";

        die "bad deltion marker" if $h->{__deleted_row} !~ /^(0|1)$/;

        if( $hasnul & $this->{fields}->{$field}->{null} ) {
            $h->{$field}->{raw}  = undef;
            $h->{$field}->{pval} = undef;
        } else {
            $h->{$field}->{raw} = substr($data, 
                $this->{fields}->{$field}->{displacement},
                $this->{fields}->{$field}->{width},
            );
            $h->{$field}->{hex} = unpack("H*", $h->{$field}->{raw});
            $h->{$field}->{pval} = $this->convert_field( $typec, $h->{$field}->{raw} );
        }
    }

    return $h;
}
# }}}
# dump_all_rows_to_main_hash {{{
sub dump_all_rows_to_main_hash {
    my $this = shift;
    my $x = 0;

    while( my $h = $this->fetchrow_hashref_by_n( $x++ ) ) {
        push @{ $this->{data} }, $h;
    }
}
# }}}
# seek_row {{{
sub seek_row {
    my $this = shift;

    $this->{cur_row} = shift;

    croak "bad row number" unless defined($this->{cur_row}) and $this->{cur_row} >= 0;
}
# }}}
# mysql_row_dump {{{
sub mysql_row_dump {
    my $this = shift;
    my $h    = $this->fetchrow_hashref_by_n( $this->{cur_row} );

    $this->{cur_row}++;

    my @fields = keys %{ $this->{fields} };
    my $fields = join(", ", map("`$_`", @fields));

    if( ref($h) eq "HASH") {
        my $values = join(", ", map(
            defined($h->{$_}->{pval}) 
                ? $this->quote($h->{$_}->{pval}) : 
                  defined($h->{$_}->{raw}) ? 
                      $this->quote( $h->{$_}->{raw} ) : 
                      "NULL", 
        @fields));

        return ($h->{__deleted_row} ? "# " : "") . "insert into $this->{table_name}($fields) values($values)";
    }

    return undef;
}
# }}}
# quote {{{
sub quote {
    my ($dbh, $str, $data_type) = @_;

    return "NULL" unless defined $str;
        unless ($data_type) {
        $str =~ s/'/''/g;       # ISO SQL2
        return "'$str'";
    }

    my $dbi_literal_quote_cache = [ {} , {} ];
    my ($prefixes, $suffixes) = @$dbi_literal_quote_cache;

    my $lp = $prefixes->{$data_type};
    my $ls = $suffixes->{$data_type}; 

    if ( ! defined $lp || ! defined $ls ) {
        my $ti = $dbh->type_info($data_type);
        $lp = $prefixes->{$data_type} = $ti ? $ti->{LITERAL_PREFIX} || "" : "'";
        $ls = $suffixes->{$data_type} = $ti ? $ti->{LITERAL_SUFFIX} || "" : "'";
    }

    return $str unless $lp || $ls; # no quoting required

    # XXX don't know what the standard says about escaping
    # in the 'general case' (where $lp != "'").
    # So we just do this and hope:
    $str =~ s/$lp/$lp$lp/g
    if $lp && $lp eq $ls && ($lp eq "'" || $lp eq '"');
    return "$lp$str$ls";
}
# }}}
# convert_field {{{
sub convert_field {
    my $this  = shift;
    my $typec = shift;
    my $val   = shift;

    if( $typec =~ m/[c]/i ) {
        $val =~ s/\s+$//;
    } elsif( $typec =~ m/[id]/i ) {
        $val = unpack("i", $val);
    } elsif( $typec =~ m/[nl]/i ) {
        # it's all set ...
    } else {
        unless( $this->{f_type_warn}->{$typec} ) {
            $this->{f_type_warn}->{$typec} = 1;
            carp "field type $typec {pval} converter incomplete, {raw} and {hex} still available";
            $val = undef;
        }
    }

    return $val;
}
# }}}
# findrows {{{
sub findrows {
    my $this     = shift;
    my $field    = shift;
    my $matchfn  = shift; ref($matchfn) eq "CODE" or croak "bad match function";
    my $matchdel = shift;

    my $fieldh = $this->{fields}->{$field} or croak "no such field ($field)";
    my $offset = $fieldh->{displacement};
    my $width  = $fieldh->{width};
    my $typec  = $fieldh->{type_c};

    my ($deleted, $data, @m) = (" ");
    for my $row (0..$this->{header}->{records}) {
        my $dpos = $this->{header}->{header_bytes} + $this->{header}->{record_bytes} * $row;
        my $opos = $dpos + $offset;

        unless( $matchdel ) { # if we match deleted, we never consider changing this to a "*"
            seek($this->{IN}, $dpos, SEEK_SET) or croak "bad seek($row, $dpos): $!";
            read($this->{IN}, $deleted, 1) == 1 or croak "bad read at $dpos for 1 byte";
        }

        if( $deleted eq " " ) { # "*" if deleted
            seek($this->{IN}, $opos, SEEK_SET) or croak "bad seek($row, $opos): $!";
            read($this->{IN}, $data, $width) == $width or croak "bad read at $opos for $width byte(s)";

            $data = $this->convert_field( $typec, $data );
            push @m, $row if $matchfn->($data);
        }
    }

    return (wantarray ? @m : \@m);
}
# }}}
# mcfindrows {{{
sub mcfindrows {
    my $this     = shift;
    my $fields   = shift; ref($fields)  eq "ARRAY" or croak "bad field names array";
    my $matchfn  = shift; ref($matchfn) eq "CODE"  or croak "bad match function";
    my $matchdel = shift;

    my @fieldh = ();
    my @offset = ();
    my @width  = ();
    my @typec  = ();

    for my $f (@$fields) {
        my $h = $this->{fields}->{$f} or croak "no such field '$f'";
        my $o = $h->{displacement};
        my $w = $h->{width};
        my $t = $h->{type_c};

        push @fieldh, $h;
        push @offset, $o;
        push @width,  $w;
        push @typec,  $t;
    }

    my ($deleted, $data, @m) = (" ");
    for my $row (0..$this->{header}->{records}) {
        my $dpos = $this->{header}->{header_bytes} + $this->{header}->{record_bytes} * $row;

        unless( $matchdel ) { # if we match deleted, we never consider changing this to a "*"
            seek($this->{IN}, $dpos, SEEK_SET) or croak "bad seek($row, $dpos): $!";
            read($this->{IN}, $deleted, 1) == 1 or croak "bad read at $dpos for 1 byte";
        }

        if( $deleted eq " " ) { # "*" if deleted
            my @data = ();

            for my $i (0..$#fieldh) {
                my $opos  = $dpos + $offset[$i];

                seek($this->{IN}, $opos, SEEK_SET) or croak "bad seek($row, $opos): $!";
                read($this->{IN}, $data, $width[$i]) == $width[$i] or croak "bad read at $opos for $width[$i] byte(s)";

                push @data, $this->convert_field( $typec[$i], $data );
            }

            push @m, $row if $matchfn->(\@data);
        }
    }

    return (wantarray ? @m : \@m);
}
# }}}

__END__

=head1 NAME

$Id: DBF.pm,v 1.45 2003/02/08 17:16:58 jettero Exp $

FoxPro::DBF - Perl extension for reading and writing foxpro .dbf files

=head1 SYNOPSIS

This module is nowhere near done yet!  I have a bazillion plans for 
features in my todo list (which is not public).  If you have any feature 
requests do send them along.  Even if I don't implement them, I'd love to 
hear from you.

=head1 example applications

These are some example applications that work (somewhat) even under
this embarassingly early release.

=head2 fp2pld.pl

    use strict;
    use FoxPro::DBF;

    my $file = "char_int.dbf";
       $file = $ARGV[0] if -f $ARGV[0];
    my $dump = $file; $dump =~ s/.dbf/.pld/i;

    my $fp = new FoxPro::DBF($file);

    $fp->dump_all_rows_to_main_hash;
    $fp->hash_dump( $dump );

=head2 fp2mysql.pl

    use strict;
    use FoxPro::DBF;

    my $file = "char_int.dbf";
       $file = $ARGV[0] if -f $ARGV[0];
    my $dump = $file; $dump =~ s/.dbf/.sql/i;

    my $fp = new FoxPro::DBF($file);

    my $temp = 0;

    open OUT, ">$dump" or die $!;

    print OUT $fp->mysql_table_drop, ";\n";
    print OUT $fp->mysql_table_create( $temp ), ";\n\n";

    $fp->seek_row( 0 );
    while( my $row = $fp->mysql_row_dump ) {
        print OUT "$row;\n";
    }

    close OUT;

=head1 other notes

    - The first row is 0, the last row is $fp->{header}->{records}-1
    - handy construct:

       for my $i (0..$fp->{header}->{records}-1) {
           my $h = $fp->fetchrow_hashref_by_n( $i );

           print "fieldname: $h->{fieldname}->{pval}\n";
       }

    - The new findrow sub:

        my $find = sub {
            my $d = shift;

            $d =~ s/\D//g;

            return ($d =~ m/$n$/ ? 1:0);
        };
        
        my @rows = $pline->findrows("phone_no", $find);

        print "@rows\n";


=head1 AUTHOR

Jettero Heller jettero@cpan.org

http://www.voltar.org

=head1 SEE ALSO

perl(1)

=cut
